@@ -22,8 +22,8 @@
 
 ;; You should have received a copy of the GNU General Public License
 ;; along with GNU Emacs; see the file COPYING.  If not, write to the
-;; Free Software Foundation, Inc., 59 Temple Place - Suite 330,
-;; Boston, MA 02111-1307, USA.
+;; Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
+;; Boston, MA 02110-1301, USA.
 
 ;;; Code:
 
@@ -32,6 +32,7 @@
 (require 'calist)
 (require 'alist)
 (require 'mime-conf)
+(require 'path-util)
 
 (eval-when-compile (require 'static))
 
@@ -69,14 +70,109 @@ buttom. Nil means don't scroll at all."
 		 (sexp :tag "Situation" 1)))
 
 (defcustom mime-view-mailcap-files
-  (let ((files '("/etc/mailcap" "/usr/etc/mailcap" "~/.mailcap")))
-    (or (member mime-mailcap-file files)
-	(setq files (cons mime-mailcap-file files)))
-    files)
+  '("~/.mailcap" "/usr/etc/mailcap" "/etc/mailcap")
   "List of mailcap files."
   :group 'mime-view
   :type '(repeat file))
 
+(defcustom mime-view-buttons-visible t
+  "Toggle visibility of MIME buttons."
+  :group 'mime-view
+  :type 'boolean)
+
+(defcustom mime-view-multipart/related-show-all-children nil
+  "When non-nil, do not hide child entities."
+  :group 'mime-view
+  :type 'boolean)
+
+(defcustom mime-view-multipart/alternative-show-all-children nil
+  "When non-nil, show hidden descendant entities's buttons in
+multipart/alternative entities."
+  :group 'mime-view
+  :type 'boolean)
+
+(defvar mime-display-multipart/multilingual-unknown-translation-type
+  "(unknown)")
+
+(defcustom mime-display-multipart/multilingual-prefered-languages
+  (mapcar (lambda (lang)
+	    (format "^%s\\(-.+\\)?" (regexp-quote (symbol-name lang))))
+	  (let ((result (get-language-info
+			 current-language-environment 'iso639-language)))
+	    (if (eq result 'en)
+		(list result)
+	      (cons result '(en)))))
+  "Specify language automatically choiced for
+multipart/multilingual entities. See docstring of
+`mime-display-multipart/multilingual' for details."
+  :group 'mime-view
+  :type '(repeat regexp))
+
+(defcustom mime-display-multipart/multilingual-translation-type-score
+  `(("original" . 2)
+    ("human" . 1)
+    ("automated" . -1)
+    (,mime-display-multipart/multilingual-unknown-translation-type . 0)
+    )
+  "Specify scores Content-Translation-Type: header fields.  When
+score is negative value, corresponding entity is not displayed
+automatically.  If field calue is missing or field does not
+exist, field value is treated as \"(unknown)\".  See docstring of
+`mime-display-multipart/multilingual' for details."
+  :group 'mime-view
+  :type '(repeat (cons string integer)))
+
+(defcustom mime-display-multipart/multilingual-interactive nil
+  "When non-nil, you are asked which language entity should be
+displayed for multipart/multilingual entity."
+  :group 'mime-view
+  :type 'boolean)
+
+(defcustom mime-view-text/html-score 3
+  "Score for text/html entity when previewer is available."
+  :group 'mime-view
+  :type 'integer)
+
+(defcustom mime-view-text/html-previewer-alist
+  (delq nil `((w3m mime-w3m-preview-text/html mime-w3m)
+	      ,(and (fboundp 'libxml-parse-html-region)
+		    '(shr mime-shr-preview-text/html mime-shr))
+	      (w3 mime-preview-text/html mime-w3)))
+  "Alist for text/html entity previewer.
+Each element is a list consists of required module, previewer function and required feature for previewer function."
+  :group 'mime-view
+  :type '(repeat (list (symbol :tag "Module")
+		       (symbol :tag "Function")
+		       (symbol :tag "Feature"))))
+
+(defcustom mime-view-text/html-previewer
+  (let ((alist mime-view-text/html-previewer-alist))
+    (while (and alist (null (module-installed-p (caar alist))))
+      (setq alist (cdr alist)))
+    (caar alist))
+  "Indicate text/html entity previewer.  Possible vaules are each car of `mime-view-text/html-previewer-alist' element or nil.  When this value is nil or preview is not available, text/html entity is displayed as if text/plain part."
+  :group 'mime-view
+  :type `(choice ,@(mapcar (lambda (elt) (list 'const (car elt)))
+			   mime-view-text/html-previewer-alist)
+		 (const :tag "Disable previewer" nil)))
+
+(defcustom mime-display-text/plain-flowed-fill-column nil
+  "Fill column for formatting flowed text."
+  :group 'mime-view
+  :type '(choice (integer :tag "Fixed value")
+		 (number :tag "ratio to window's width")
+		 (sexp :tag "S-expression")
+		 (const nil :tag "Use fill-column's value")))
+
+(defcustom mime-pgp-verify-when-preview t
+  "When non-nil, verify signed part while viewing."
+  :group 'mime-view
+  :type 'boolean)
+
+(defcustom mime-pgp-decrypt-when-preview nil
+  "When non-nil, decrypt encrypted part while viewing."
+  :group 'mime-view
+  :type 'boolean)
 
 ;;; @ in raw-buffer (representation space)
 ;;;
@@ -120,8 +216,7 @@ If optional argument RECURSIVE is non-nil and current buffer has
 mime-mother-buffer, it returns original major-mode of the
 mother-buffer."
   (if (and recursive mime-mother-buffer)
-      (save-excursion
-	(set-buffer mime-mother-buffer)
+      (with-current-buffer mime-mother-buffer
 	(mime-preview-original-major-mode recursive)
 	)
     (cdr (assq 'major-mode
@@ -345,8 +440,8 @@ mother-buffer."
 
 (defun mime-view-entity-title (entity)
   (or (mime-entity-read-field entity 'Content-Description)
-      (mime-entity-read-field entity 'Subject)
       (mime-entity-filename entity)
+      (mime-entity-read-field entity 'Subject)
       ""))
 
 (defvar mime-preview-situation-example-list nil)
@@ -423,33 +518,34 @@ mother-buffer."
 	  mime-acting-situation-example-list)
       (let ((file mime-situation-examples-file)
 	    print-length print-level)
-	(with-temp-buffer
-	  (insert ";;; " (file-name-nondirectory file) "\n")
-	  (insert "\n;; This file is generated automatically by "
-		  mime-view-version "\n\n")
-	  (insert ";;; Code:\n\n")
-	  (if mime-preview-situation-example-list
-	      (pp `(setq mime-preview-situation-example-list
-			 ',mime-preview-situation-example-list)
-		  (current-buffer)))
-	  (if mime-acting-situation-example-list
-	      (pp `(setq mime-acting-situation-example-list
-			 ',mime-acting-situation-example-list)
-		  (current-buffer)))
-	  (insert "\n;;; "
-		  (file-name-nondirectory file)
-		  " ends here.\n")
-	  (static-cond
-	   ((boundp 'buffer-file-coding-system)
-	    (setq buffer-file-coding-system
-		  mime-situation-examples-file-coding-system))
-	   ((boundp 'file-coding-system)
-	    (setq file-coding-system
-		  mime-situation-examples-file-coding-system)))
-	  ;; (setq buffer-file-coding-system
-	  ;;       mime-situation-examples-file-coding-system)
-	  (setq buffer-file-name file)
-	  (save-buffer)))))
+        (when file
+          (with-temp-buffer
+            (insert ";;; " (file-name-nondirectory file) "\n")
+            (insert "\n;; This file is generated automatically by "
+                    mime-view-version "\n\n")
+            (insert ";;; Code:\n\n")
+            (if mime-preview-situation-example-list
+                (pp `(setq mime-preview-situation-example-list
+                           ',mime-preview-situation-example-list)
+                    (current-buffer)))
+            (if mime-acting-situation-example-list
+                (pp `(setq mime-acting-situation-example-list
+                           ',mime-acting-situation-example-list)
+                    (current-buffer)))
+            (insert "\n;;; "
+                    (file-name-nondirectory file)
+                    " ends here.\n")
+            (static-cond
+             ((boundp 'buffer-file-coding-system)
+              (setq buffer-file-coding-system
+                    mime-situation-examples-file-coding-system))
+             ((boundp 'file-coding-system)
+              (setq file-coding-system
+                    mime-situation-examples-file-coding-system)))
+            ;; (setq buffer-file-coding-system
+            ;;       mime-situation-examples-file-coding-system)
+            (setq buffer-file-name file)
+            (save-buffer))))))
 
 (add-hook 'kill-emacs-hook 'mime-save-situation-examples)
 
@@ -599,24 +695,27 @@ mother-buffer."
 		  )))
 	    )
 	   (t
-	    (let ((media-type (mime-entity-media-type entity))
-		  (media-subtype (mime-entity-media-subtype entity))
-		  (charset (cdr (assoc "charset" params)))
-		  (encoding (mime-entity-encoding entity)))
+	    (let* ((charset (cdr (assoc "charset" params)))
+		   (encoding (mime-entity-encoding entity))
+		   (language (mime-entity-read-field
+			      entity "Content-Language"))
+		   (rest (format " <%s/%s%s%s%s>"
+				 (mime-entity-media-type entity)
+				 (mime-entity-media-subtype entity)
+				 (if language
+				     (concat " (" language ")")
+				   "")
+				 (if charset
+				     (concat "; " charset)
+				   "")
+				 (if encoding
+				     (concat " (" encoding ")")
+				   ""))))
 	      (concat
 	       num " " subject
-	       (let ((rest
-		      (format " <%s/%s%s%s>"
-			      media-type media-subtype
-			      (if charset
-				  (concat "; " charset)
-				"")
-			      (if encoding
-				  (concat " (" encoding ")")
-				""))))
-		 (if (>= (+ (current-column)(length rest))(window-width))
-		     "\n\t")
-		 rest)))
+	       (if (>= (+ (current-column)(length rest))(window-width))
+		   "\n\t")
+	       rest))
 	    )))
      (function mime-preview-play-current-entity))
     ))
@@ -667,6 +766,19 @@ Each elements are regexp of field-name.")
 (define-calist-field-match-method
   'body #'mime-calist::field-match-method-as-default-rule)
 
+(defun mime-calist::field-match-method-ignore-case (calist
+						    field-type field-value)
+  (let ((s-field (assoc field-type calist)))
+    (cond ((null s-field)
+	   (cons (cons field-type field-value) calist))
+	  ((eq field-value t)
+	   calist)
+	  ((string= (downcase (cdr s-field)) (downcase field-value))
+	   calist))))
+
+(define-calist-field-match-method
+  'access-type #'mime-calist::field-match-method-ignore-case)
+
 
 (defvar mime-preview-condition nil
   "Condition-tree about how to display entity.")
@@ -727,6 +839,12 @@ Each elements are regexp of field-name.")
    (body . visible)
    (body-presentation-method . mime-display-text/richtext)))
 
+(ctree-set-calist-strictly
+ 'mime-preview-condition
+ '((type . text)(subtype . html)
+   (body . visible)
+   (body-presentation-method . mime-display-text/html)))
+
 (autoload 'mime-display-application/x-postpet "postpet")
 
 (ctree-set-calist-strictly
@@ -753,6 +871,12 @@ Each elements are regexp of field-name.")
    (body . visible)
    (body-presentation-method . mime-display-multipart/related)))
 
+(ctree-set-calist-strictly
+ 'mime-preview-condition
+ '((type . multipart)(subtype . multilingual)
+   (body . visible)
+   (body-presentation-method . mime-display-multipart/multilingual)))
+
 (ctree-set-calist-strictly
  'mime-preview-condition
  '((type . multipart)(subtype . t)
@@ -769,59 +893,160 @@ Each elements are regexp of field-name.")
  'mime-preview-condition
  '((type . message)(subtype . rfc822)
    (body . visible)
-   (body-presentation-method . mime-display-multipart/mixed)
-   (childrens-situation (header . visible)
-			(entity-button . invisible))))
+   (body-presentation-method . mime-display-message/rfc822)
+   (childrens-situation (header . visible))))
 
 (ctree-set-calist-strictly
  'mime-preview-condition
  '((type . message)(subtype . news)
    (body . visible)
-   (body-presentation-method . mime-display-multipart/mixed)
-   (childrens-situation (header . visible)
-			(entity-button . invisible))))
+   (body-presentation-method . mime-display-message/rfc822)
+   (childrens-situation (header . visible))))
 
 
 ;;; @@@ entity presentation
 ;;;
 
+(defun mime-display-insert-text-content (entity)
+  (condition-case signal
+      (progn (mime-insert-text-content entity)
+	     (run-hooks 'mime-text-decode-hook)
+	     t)
+    (error   (let ((point (point)))
+	       (insert (format "This entity can't be decoded.  %s"
+			       (or (cadr signal) "")))
+	       (add-text-properties point (point) '(face highlight)))
+	     (insert "\nHere is original data.\n\n")
+	     (mime-insert-entity-body entity)
+	     nil)))
+
+(defun mime-display-text/plain-flowed-parse-line ()
+  (cons (point)
+	(cond
+	 ;; End of buffer
+	 ((eobp) nil)
+	 ;; Signature separator line
+	 ((looking-at "\\(>+\\)? ?\\(-- \\)$")
+	  (list (length (match-string 1)) 'hard (match-beginning 2)))
+	 ;; Quoted line
+	 ((looking-at "\\(>+\\) ?\\(.*?\\)\\( ?\\)$")
+	  (list (length (match-string 1))
+		(if (zerop (length (match-string 3))) t nil)
+		(match-beginning 2)))
+	 ;; Stuffed or normal line
+	 ((looking-at " ?\\(.*?\\)\\( ?\\)$")
+	  (list 0 (if (zerop (length (match-string 2))) t nil)
+		(match-beginning 1)))
+	 )))
+
+(defun mime-display-text/plain-flowed (&optional buffer delete-space)
+  (with-current-buffer (or (current-buffer) buffer)
+    (goto-char (point-max))
+    (unless (eq (char-before) ?\n)
+      (insert ?\n))
+    (goto-char (point-min))
+    (let ((fill-column
+	   (cond
+	    ((and (integerp mime-display-text/plain-flowed-fill-column)
+		  (< mime-display-text/plain-flowed-fill-column 1))
+	     (+ (window-width) mime-display-text/plain-flowed-fill-column))
+	    ((integerp mime-display-text/plain-flowed-fill-column)
+	     mime-display-text/plain-flowed-fill-column)
+	    ((numberp mime-display-text/plain-flowed-fill-column)
+	     (floor
+	      (* (window-width) mime-display-text/plain-flowed-fill-column)))
+	    (mime-display-text/plain-flowed-fill-column
+	     (eval mime-display-text/plain-flowed-fill-column))
+	    (t fill-column)))
+	  (line (mime-display-text/plain-flowed-parse-line))
+	  beg-flow depth next point fill-prefix adaptive-fill-mode)
+      (setq delete-space (if delete-space -2 -1))
+      (while (cdr line)
+	(unless (eq (point) (car line))
+	  (setcar (cdr (cddr line)) (+ (nth 3 line) (- (point) (car line))))
+	  (setcar line (point)))
+	(forward-line)
+	(setq next (mime-display-text/plain-flowed-parse-line))
+	(when (or (null (cdr next))
+		  (null (eq (nth 1 line) (nth 1 next)))
+		  (eq (nth 2 next) 'hard))
+	  (setcar (cddr line) t))
+	(if beg-flow
+	    ;; Following flowed line.
+	    (progn
+	      (delete-region (+ (car line) delete-space) (nth 3 line))
+	      (when (nth 2 line)
+		;; Fixed line
+		(setq fill-prefix (unless (zerop depth)
+				    (concat (make-string depth ?>) " ")))
+		(fill-region beg-flow (point) 'left t)
+		(setq beg-flow nil)))
+	  ;; Following fixed line.
+	  (if (zerop (nth 1 line))
+	      ;; Remove stuffed space
+	      (delete-region (car line) (nth 3 line))
+	    (when (eq (+ (car line) (nth 1 line)) (nth 3 line))
+	      ;; Insert stuffing space for quoted text
+	      (setq point (point))
+	      (goto-char (nth 3 line))
+	      (insert 32)
+	      (goto-char (1+ point))))
+	  (unless (nth 2 line)
+	    ;; Beginnig of flowed text
+	    (setq beg-flow (car line)
+		  depth (nth 1 line))))
+	(setq line next)
+	))))
+
 (defun mime-display-text/plain (entity situation)
   (save-restriction
     (narrow-to-region (point-max)(point-max))
-    (condition-case nil
-	(mime-insert-text-content entity)
-      (error (progn
-	       (message "Can't decode current entity.")
-	       (sit-for 1))))
-    (run-hooks 'mime-text-decode-hook)
-    (goto-char (point-max))
-    (if (not (eq (char-after (1- (point))) ?\n))
+    (when (mime-display-insert-text-content entity)
+      (when (not (eq (char-before (point-max)) ?\n))
+	(goto-char (point-max))
 	(insert "\n")
-      )
-    (mime-add-url-buttons)
-    (run-hooks 'mime-display-text/plain-hook)
-    ))
+	)
+      (when (equal (downcase (or (cdr (assoc "format" situation)) ""))
+		   "flowed")
+	(mime-display-text/plain-flowed
+	 nil (equal (downcase (or (cdr (assoc "delsp" situation)) ""))
+		    "yes")))
+      (mime-add-url-buttons)
+      (run-hooks 'mime-display-text/plain-hook)
+      )))
 
 (defun mime-display-text/richtext (entity situation)
   (save-restriction
     (narrow-to-region (point-max)(point-max))
-    (mime-insert-text-content entity)
-    (run-hooks 'mime-text-decode-hook)
-    (let ((beg (point-min)))
-      (remove-text-properties beg (point-max) '(face nil))
-      (richtext-decode beg (point-max))
+    (when (mime-display-insert-text-content entity)
+      (remove-text-properties (point-min) (point-max) '(face nil))
+      (richtext-decode (point-min) (point-max))
       )))
 
 (defun mime-display-text/enriched (entity situation)
   (save-restriction
     (narrow-to-region (point-max)(point-max))
-    (mime-insert-text-content entity)
-    (run-hooks 'mime-text-decode-hook)
-    (let ((beg (point-min)))
-      (remove-text-properties beg (point-max) '(face nil))
-      (enriched-decode beg (point-max))
+    (when (mime-display-insert-text-content entity)
+      (remove-text-properties (point-min) (point-max) '(face nil))
+      (enriched-decode (point-min) (point-max))
       )))
 
+(defun mime-display-text/html-previewer-params ()
+  (and mime-view-text/html-previewer
+       (or (assq mime-view-text/html-previewer
+		 mime-view-text/html-previewer-alist)
+	   ;; For compatibility with mime-setup-enable-inline-html.
+	   (assq 'w3 mime-view-text/html-previewer-alist))))
+
+(defun mime-display-text/html (entity situation)
+  (let ((list (mime-display-text/html-previewer-params)))
+    (if (and list
+	     (require (nth 2 list) nil t)
+	     (fboundp (nth 1 list)))
+	(funcall (nth 1 list) entity situation)
+      ;; text/html entity previewer is not available.
+      (mime-display-text/plain entity situation))))
+
 
 (defvar mime-view-announcement-for-message/partial
   (if (and (>= emacs-major-version 19) window-system)
@@ -847,6 +1072,19 @@ Each elements are regexp of field-name.")
 		     #'mime-preview-play-current-entity)
     ))
 
+(defun mime-display-message/rfc822 (entity situation)
+  (let ((child (car (mime-entity-children entity)))
+	(default-situation
+	  (copy-alist
+	   (delq nil (cons (assq 'major-mode situation)
+			   (cdr (assq 'childrens-situation situation)))))))
+    (mime-display-entity
+     child nil
+     (if (memq (mime-entity-media-type child)
+	       '(text multipart nil))
+	 (put-alist 'entity-button 'invisible default-situation)
+       (put-alist 'button-position 'after default-situation)))))
+
 (defun mime-display-multipart/mixed (entity situation)
   (let ((children (mime-entity-children entity))
 	(original-major-mode-cell (assq 'major-mode situation))
@@ -860,13 +1098,19 @@ Each elements are regexp of field-name.")
       (setq children (cdr children))
       )))
 
+(defvar mime-view-entity-lowest-score -1)
+
 (defcustom mime-view-type-subtype-score-alist
   '(((text . enriched) . 3)
     ((text . richtext) . 2)
     ((text . plain)    . 1)
-    (t . 0))
+    ((text . html)     . mime-view-text/html-entity-score)
+    (multipart . mime-view-multipart-entity-score)
+    )
   "Alist MEDIA-TYPE vs corresponding score.
-MEDIA-TYPE must be (TYPE . SUBTYPE), TYPE or t.  t means default."
+MEDIA-TYPE must be (TYPE . SUBTYPE), TYPE or t.  t means default.
+If MEDIA-TYPE does not have corresponding score, `mime-view-entity-lowest-score' is used.
+Score is integer or function or variable.  The function receives entity and returns integer."
   :group 'mime-view
   :type '(repeat (cons (choice :tag "Media-Type"
 			       (cons :tag "Type/Subtype"
@@ -874,60 +1118,99 @@ MEDIA-TYPE must be (TYPE . SUBTYPE), TYPE or t.  t means default."
 				     (symbol :tag "Subtype"))
 			       (symbol :tag "Type")
 			       (const :tag "Default" t))
-		       integer)))
+		       (choice (integer :tag "score")
+			       (function :tag "function")
+			       (variable :tag "variable")))))
+
+(defun mime-view-entity-score (entity &optional situation)
+  (or situation (setq situation (mime-entity-situation entity)))
+  (let ((score
+	 (cdr
+	  (or (assoc (cons (cdr (assq 'type situation))
+			   (cdr (assq 'subtype situation)))
+		     mime-view-type-subtype-score-alist)
+	      (assq (cdr (assq 'type situation))
+		    mime-view-type-subtype-score-alist)
+	      (assq t mime-view-type-subtype-score-alist)
+	      ))))
+    (cond
+     ((functionp score)
+      (setq score (funcall score entity)))
+     ((and (symbolp score) (boundp score))
+      (setq score (symbol-value score))))
+    (if (numberp score)
+	score
+      mime-view-entity-lowest-score)
+    ))
+
+(defun mime-view-multipart-entity-score (entity)
+  (apply 'max (or (mapcar 'mime-view-entity-score
+			  (mime-entity-children entity))
+		  (list (or (assq t mime-view-type-subtype-score-alist)
+			    mime-view-entity-lowest-score)))))
+
+(defun mime-view-text/html-entity-score (entity)
+  ;; Module loading is done in mime-display-text/html.
+  ;; So, availability is not checked here.
+  (if (mime-display-text/html-previewer-params)
+      mime-view-text/html-score
+    mime-view-entity-lowest-score))
+
+(defun mime-view-multipart-descendant-button (entity situation)
+  (let ((default-situation
+	  (delq nil (cons (assq 'major-mode situation)
+			  (cdr (assq 'childrens-situation situation))))))
+    (mapc
+     (lambda (child)
+       (setq situation (copy-alist (mime-find-entity-preview-situation
+				    child default-situation)))
+       (mime-display-entity
+	child nil (if (eq (cdr (assq 'type situation)) 'multipart)
+		      (put-alist 'body-presentation-method
+				 'mime-view-multipart-descendant-button
+				 situation)
+		    (put-alist 'body 'invisible situation))))
+     (mime-entity-children entity))))
 
 (defun mime-display-multipart/alternative (entity situation)
-  (let* ((children (mime-entity-children entity))
-	 (original-major-mode-cell (assq 'major-mode situation))
-	 (default-situation
-	   (cdr (assq 'childrens-situation situation)))
-	 (i 0)
-	 (p 0)
-	 (max-score 0)
-	 situations)
-    (if original-major-mode-cell
-	(setq default-situation
-	      (cons original-major-mode-cell default-situation)))
-    (setq situations
-	  (mapcar (function
-		   (lambda (child)
-		     (let ((situation
-			    (mime-find-entity-preview-situation
-			     child default-situation)))
-		       (if (cdr (assq 'body-presentation-method situation))
-			   (let ((score
-				  (cdr
-				   (or (assoc
-					(cons
-					 (cdr (assq 'type situation))
-					 (cdr (assq 'subtype situation)))
-					mime-view-type-subtype-score-alist)
-				       (assq
-					(cdr (assq 'type situation))
-					mime-view-type-subtype-score-alist)
-				       (assq
-					t
-					mime-view-type-subtype-score-alist)
-				       ))))
-			     (if (> score max-score)
-				 (setq p i
-				       max-score score)
-			       )))
-		       (setq i (1+ i))
-		       situation)
-		     ))
-		  children))
-    (setq i 0)
-    (while children
-      (let ((child (car children))
-	    (situation (car situations)))
-	(mime-display-entity child (if (= i p)
-				       situation
-				     (put-alist 'body 'invisible
-						(copy-alist situation)))))
-      (setq children (cdr children)
-	    situations (cdr situations)
-	    i (1+ i)))))
+  (let ((original-major-mode-cell (assq 'major-mode situation))
+	(default-situation
+	  (cdr (assq 'childrens-situation situation)))
+	(max-score 0)
+	p pairs
+	child-situation score)
+    (when original-major-mode-cell
+      (setq default-situation
+	    (cons original-major-mode-cell default-situation)))
+    (setq pairs
+	  (mapcar
+	   (lambda (child)
+	     (setq child-situation
+		   (mime-find-entity-preview-situation
+		    child default-situation))
+	     (when (cdr (assq 'body-presentation-method child-situation))
+	       (setq score (mime-view-entity-score child child-situation))
+	       (when (>= score max-score)
+		 (setq p child
+		       max-score score)
+		 ))
+	     (cons child child-situation))
+	   (mime-entity-children entity)))
+    (or p (setq p (caar pairs)))
+    (mapc (lambda (pair)
+	    (mime-display-entity
+	     (car pair) nil
+	     (cond
+	      ((eq p (car pair))
+	       (cdr pair))
+	      ((and mime-view-multipart/alternative-show-all-children
+		    (eq (cdr (assq 'type (cdr pair))) 'multipart))
+	       (put-alist 'body-presentation-method
+			  'mime-view-multipart-descendant-button
+			  (copy-alist (cdr pair))))
+	      (t (put-alist 'body 'invisible (copy-alist (cdr pair)))))))
+	  pairs)
+    ))
 
 (defun mime-display-multipart/related (entity situation)
   (let* ((param-start (mime-parse-msg-id
@@ -944,7 +1227,127 @@ MEDIA-TYPE must be (TYPE . SUBTYPE), TYPE or t.  t means default."
     (if original-major-mode-cell
 	(setq default-situation
 	      (cons original-major-mode-cell default-situation)))
-    (mime-display-entity start nil default-situation)))
+    (mapc
+     (lambda (child)
+       (cond
+	((eq start child)
+	 (mime-display-entity start nil default-situation))
+	(mime-view-multipart/related-show-all-children
+	 (let ((child-situation (copy-alist (mime-find-entity-preview-situation
+					     child default-situation))))
+	   (mime-display-entity
+	    child nil
+	    (if (eq (mime-entity-media-type child) 'multipart)
+		(put-alist 'body-presentation-method
+			   'mime-view-multipart-descendant-button
+			   child-situation)
+	      (put-alist 'body 'invisible child-situation)))))))
+     (mime-entity-children entity))))
+
+(defun mime-display-multipart/multilingual-select-interactively (pairs)
+  (let (count counts result elt)
+    (setq pairs
+	  (dolist (pair pairs (nreverse result))
+	    (when (caar pair)
+	      (setq elt (format "%s/%s"
+				(or (caar pair) "")
+				(or (cdar pair)
+				    mime-display-multipart/multilingual-unknown-translation-type)))
+	      (if (setq count (assoc elt counts))
+		  (setq elt
+			(format "%s:%d" elt (setcdr count (1+ (cdr count)))))
+		(setq counts (cons (cons elt 1) counts)))
+	      (setq result (cons (cons elt (cdr pair)) result)))))
+    (when (> (length pairs) 0)
+      (cdr (assoc
+	    (completing-read
+	     "Which language is displayed for this multilingual message? "
+	     (mapcar 'car pairs) nil t nil nil (caar pairs))
+	    pairs)))))
+
+(defun mime-display-multipart/multilingual-select-automatically (pairs)
+  (let ((max-score '(0 . 0))
+	(case-fold-search t)
+	score choice first zxx)
+    (dolist (pair pairs)
+      (when (caar pair)
+	(unless first
+	  ;; The first language message part
+	  (setq first (cdr pair)))
+	(when (string-match "^zxx$" (caar pair))
+	  ;; The language-independent message part
+	  (setq zxx (cdr pair)))
+	(setq score
+	      (cons
+	       (let ((langs
+		      mime-display-multipart/multilingual-prefered-languages))
+		 (catch 'done
+		   (while langs
+		     (when (string-match (car langs) (caar pair))
+		       (throw 'done (length langs)))
+		     (setq langs (cdr langs)))
+		   -1))
+	       (or
+		(cdr
+		 (assoc
+		  (or (cdar pair)
+		      mime-display-multipart/multilingual-unknown-translation-type)
+		  mime-display-multipart/multilingual-translation-type-score))
+		-1)))
+	(when (or (> (cdr score) (cdr max-score))
+		  (and (eq (cdr score) (cdr max-score))
+		       (> (car score) (car max-score))))
+	  (setq max-score score)
+	  (setq choice (cdr pair)))))
+    (or choice zxx first)))
+
+(defun mime-display-multipart/multilingual (entity situation)
+  "MIME-View mode preview method for multipart/multilingual entity.
+When `mime-display-multipart/multilingual-interactive' is nil, select child entity to display automatically.
+Automatic selection algorithm is below.
+
+1. Select highest score entity calculated from
+Content-Translation-Type: field and
+`mime-display-multipart/multilingual-translation-type-score'.
+But if score is negative, never selected.
+2. If there are multiple highest score entities, select entities
+whose Content-Language: field values matches former element of
+`mime-display-multipart/multilingual-prefered-languages'.  If not
+matched, never selected.
+3. If no entity is selected, select the language-independent
+part (if exist) or the first language message part."
+  (let ((default-situation
+	  (delq nil (cons (assq 'major-mode situation)
+			  (cdr (assq 'childrens-situation situation)))))
+	child-situation score choice pairs)
+    (setq pairs
+	  (mapcar
+	   (lambda (child)
+	     (cons (cons
+		    (cdr (assq 'atom
+			       (std11-lexical-analyze
+				(mime-entity-read-field
+				 child "Content-Language"))))
+		    (cdr (assq 'atom
+			       (std11-lexical-analyze
+				(mime-entity-fetch-field
+				 child "Content-Translation-Type")))))
+		   child))
+	   (mime-entity-children entity)))
+    (setq choice
+	  (if mime-display-multipart/multilingual-interactive
+	      (mime-display-multipart/multilingual-select-interactively pairs)
+	    (mime-display-multipart/multilingual-select-automatically pairs)))
+    (mapc (lambda (child)
+	    (mime-display-entity
+	     child nil
+	     (if (eq choice child)
+		 default-situation
+	       (put-alist 'body 'invisible
+			  (copy-alist (mime-find-entity-preview-situation
+				       child default-situation))))))
+	  (mime-entity-children entity))
+    ))
 
 ;;; @ acting-condition
 ;;;
@@ -954,7 +1357,11 @@ MEDIA-TYPE must be (TYPE . SUBTYPE), TYPE or t.  t means default."
 
 (defun mime-view-read-mailcap-files (&optional files)
   (or files
-      (setq files mime-view-mailcap-files))
+      (setq files
+	    (if mime-mailcap-file
+		(cons mime-mailcap-file
+		      (remove mime-mailcap-file mime-view-mailcap-files))
+	      mime-view-mailcap-files)))
   (let (entries file)
     (while files
       (setq file (car files))
@@ -1069,68 +1476,78 @@ MEDIA-TYPE must be (TYPE . SUBTYPE), TYPE or t.  t means default."
 ;;; @ buffer setup
 ;;;
 
+(defun mime-display-entity-visible-p (elements situation &optional default)
+  (catch 'done
+    (let (result)
+      (while elements
+	(when (setq result (cdr (assq (car elements) situation)))
+	  (unless (memq result '(visible invisible))
+	    (setq result (cond ((functionp result)
+				(funcall result situation))
+			       ((and (symbolp result) (boundp result))
+				(symbol-value result))
+			       (t default))))
+	  ;; Non-nil values other than invisible are treated as
+	  ;; visible.
+	  (throw 'done (if (memq result '(invisible nil))
+			   nil t)))
+	(setq elements (cdr elements))))
+    default))
+
 (defun mime-display-entity (entity &optional situation
 				   default-situation preview-buffer)
   (or preview-buffer
       (setq preview-buffer (current-buffer)))
-  (let* (e nb ne nhb nbb)
+  (let* (nb ne nbb)
     (in-calist-package 'mime-view)
     (or situation
 	(setq situation
 	      (mime-find-entity-preview-situation entity default-situation)))
-    (let ((button-is-invisible
-	   (eq (cdr (or (assq '*entity-button situation)
-			(assq 'entity-button situation)))
-	       'invisible))
+    (let ((button-is-visible
+	   (and mime-view-buttons-visible
+		(mime-display-entity-visible-p
+		 '(*entity-button entity-button) situation t)))
+	  (button-position (cdr (assq 'button-position situation)))
 	  (header-is-visible
-	   (eq (cdr (or (assq '*header situation)
-			(assq 'header situation)))
-	       'visible))
+	   (mime-display-entity-visible-p '(*header header) situation))
 	  (body-is-visible
-	   (eq (cdr (or (assq '*body situation)
-			(assq 'body situation)))
-	       'visible))
+	   (mime-display-entity-visible-p '(*body body) situation))
 	  (children (mime-entity-children entity)))
       (set-buffer preview-buffer)
       (setq nb (point))
       (narrow-to-region nb nb)
-      (or button-is-invisible
-          ;; (if (mime-view-entity-button-visible-p entity)
-	  (mime-view-insert-entity-button entity)
-          ;;   )
-	  )
       (if header-is-visible
 	  (let ((header-presentation-method
 		 (or (cdr (assq 'header-presentation-method situation))
 		     (cdr (assq (cdr (assq 'major-mode situation))
 				mime-header-presentation-method-alist)))))
-	    (setq nhb (point))
 	    (if header-presentation-method
 		(funcall header-presentation-method entity situation)
 	      (mime-insert-header entity
 				  mime-view-ignored-field-list
 				  mime-view-visible-field-list))
 	    (run-hooks 'mime-display-header-hook)
-	    (put-text-property nhb (point-max) 'mime-view-entity-header entity)
+	    (put-text-property nb (point-max) 'mime-view-entity-header entity)
 	    (goto-char (point-max))
 	    (insert "\n")))
+      (when button-is-visible
+	(unless (eq button-position 'after)
+	  (goto-char (point-min)))
+	(mime-view-insert-entity-button entity)
+	(goto-char (point-max)))
       (setq nbb (point))
       (unless children
-	(if body-is-visible
-	    (let ((body-presentation-method
-		   (cdr (assq 'body-presentation-method situation))))
-	      (if (functionp body-presentation-method)
-		  (funcall body-presentation-method entity situation)
-		(mime-display-text/plain entity situation)))
-	  (when button-is-invisible
-	    (goto-char (point-max))
-	    (mime-view-insert-entity-button entity)
-	    )
-	  (unless header-is-visible
-	    (goto-char (point-max))
-	    (insert "\n"))
-	  ))
-      (setq ne (point-max))
+	(when body-is-visible
+	  (let ((body-presentation-method
+		 (cdr (assq 'body-presentation-method situation))))
+	    (if (functionp body-presentation-method)
+		(funcall body-presentation-method entity situation)
+	      (mime-display-text/plain entity situation))))
+	(goto-char (point-max))
+	;; Insert LF if needed.
+	(unless (eq (char-before) 10) (insert 10)))
+      (setq ne (or (next-single-property-change nb 'mime-view-entity)
+		   (point-max)))
       (widen)
       (put-text-property nb ne 'mime-view-entity entity)
       (put-text-property nb ne 'mime-view-situation situation)
@@ -1293,13 +1710,13 @@ MEDIA-TYPE must be (TYPE . SUBTYPE), TYPE or t.  t means default."
 	   (define-key mime-view-mode-map [menu-bar mime-view]
 	     (cons mime-view-menu-title
 		   (make-sparse-keymap mime-view-menu-title)))
-	   (mapcar (function
-		    (lambda (item)
-		      (define-key mime-view-mode-map
-			(vector 'menu-bar 'mime-view (car item))
-			(cons (nth 1 item)(nth 2 item)))
-		      ))
-		   (reverse mime-view-menu-list))
+	   (mapc (function
+		  (lambda (item)
+		    (define-key mime-view-mode-map
+		      (vector 'menu-bar 'mime-view (car item))
+		      (cons (nth 1 item)(nth 2 item)))
+		    ))
+		 (reverse mime-view-menu-list))
 	   ))
     ;; (run-hooks 'mime-view-define-keymap-hook)
     mime-view-mode-map))
@@ -1311,8 +1728,7 @@ MEDIA-TYPE must be (TYPE . SUBTYPE), TYPE or t.  t means default."
   "Clear mime-echo buffer and delete window for it."
   (let ((buf (get-buffer mime-echo-buffer-name)))
     (if buf
-	(save-excursion
-	  (set-buffer buf)
+	(with-current-buffer buf
 	  (erase-buffer)
 	  (let ((win (get-buffer-window buf)))
 	    (if win
@@ -1361,11 +1777,16 @@ non-nil, DEFAULT-KEYMAP-OR-FUNCTION is ignored.  If it is nil,
       (setq mime-preview-original-window-configuration win-conf)
       (setq major-mode 'mime-view-mode)
       (setq mode-name "MIME-View")
-      (mime-display-entity message nil
-			   `((entity-button . invisible)
-			     (header . visible)
-			     (major-mode . ,original-major-mode))
-			   preview-buffer)
+      ;; Do not hide button when first entity is
+      ;; neither text nor multipart.
+      (mime-display-entity
+       message nil `(,(if (memq (mime-entity-media-type message)
+				'(text multipart nil))
+			  '(entity-button . invisible)
+			'(button-position . after))
+		     (header . visible)
+		     (major-mode . ,original-major-mode))
+       preview-buffer)
       (use-local-map
        (or keymap
 	   (if default-keymap-or-function
@@ -1401,6 +1822,8 @@ message.  It must be nil, `binary' or `cooked'.  If it is nil,
       (setq raw-buffer (current-buffer)))
   (or representation-type
       (setq representation-type
+	    ;; Do not use `with-current-buffer'.
+	    ;; raw-buffer may be the current buffer.
 	    (save-excursion
 	      (set-buffer raw-buffer)
 	      (cdr (or (assq major-mode mime-raw-representation-type-alist)
@@ -1479,11 +1902,13 @@ If WITH-CHILDREN, refer boundary surrounding current part and its branches."
   (let (entity
 	p-beg p-end
 	entity-node-id len)
-    (while (null (setq entity
-		       (get-text-property (point) 'mime-view-entity)))
+    (while (and
+	    (null (setq entity
+			(get-text-property (point) 'mime-view-entity)))
+	    (> (point) (point-min)))
       (backward-char))
     (setq p-beg (previous-single-property-change (point) 'mime-view-entity))
-    (setq entity-node-id (mime-entity-node-id entity))
+    (setq entity-node-id (and entity (mime-entity-node-id entity)))
     (setq len (length entity-node-id))
     (cond ((null p-beg)
 	   (setq p-beg
@@ -1559,13 +1984,13 @@ It decodes current entity to call internal or external method as
 It calls following-method selected from variable
 `mime-preview-following-method-alist'."
   (interactive)
-  (let ((entity (mime-preview-find-boundary-info t))
-	p-beg p-end
-	pb-beg)
-    (setq p-beg (aref entity 0)
-	  p-end (aref entity 1)
-	  entity (aref entity 2))
-    (if (get-text-property p-beg 'mime-view-entity-body)
+  (let* ((boundary-info (mime-preview-find-boundary-info t))
+	 (p-beg (aref boundary-info 0))
+	 (p-end (aref boundary-info 1))
+	 (entity (aref boundary-info 2))
+	 pb-beg)
+    (if (or (get-text-property p-beg 'mime-view-entity-body)
+	    (null entity))
 	(setq pb-beg p-beg)
       (setq pb-beg
 	    (next-single-property-change
@@ -1573,12 +1998,14 @@ It calls following-method selected from variable
 	     (or (next-single-property-change p-beg 'mime-view-entity)
 		 p-end))))
     (let* ((mode (mime-preview-original-major-mode 'recursive))
-	   (entity-node-id (mime-entity-node-id entity))
+	   (entity-node-id (and entity (mime-entity-node-id entity)))
 	   (new-name
 	    (format "%s-%s" (buffer-name) (reverse entity-node-id)))
 	   new-buf
 	   (the-buf (current-buffer))
 	   fields)
+      ;; Do not use `with-current-buffer'.  Inner save-excursion(),
+      ;; the current buffer may be accessed.
       (save-excursion
 	(set-buffer (setq new-buf (get-buffer-create new-name)))
 	(erase-buffer)
@@ -1586,7 +2013,8 @@ It calls following-method selected from variable
 	(insert-buffer-substring the-buf pb-beg p-end)
 	(goto-char (point-min))
 	(let ((current-entity
-	       (if (and (eq (mime-entity-media-type entity) 'message)
+	       (if (and entity
+			(eq (mime-entity-media-type entity) 'message)
 			(eq (mime-entity-media-subtype entity) 'rfc822))
 		   (car (mime-entity-children entity))
 		 entity)))
@@ -1607,8 +2035,7 @@ It calls following-method selected from variable
 	    (setq field-name (car rest))
 	    (or (std11-field-body field-name)
 		(progn
-		  (save-excursion
-		    (set-buffer the-buf)
+		  (with-current-buffer the-buf
 		    (let ((entity (when mime-mother-buffer
 				    (set-buffer mime-mother-buffer)
 				    (get-text-property (point)
@@ -1618,7 +2045,7 @@ It calls following-method selected from variable
 						   entity field-name))))
 			(setq entity (mime-entity-parent entity)))))
 		  (if ret
-		      (insert (concat field-name ": " ret "\n"))
+		      (insert field-name ": " ret "\n")
 		    )))
 	    (setq rest (cdr rest))
 	    ))
@@ -1817,9 +2244,9 @@ If LINES is negative, scroll up LINES lines."
 	  (display)
 	  (t
 	   (setq display
-		 (eq (cdr (or (assq sym situation)
-			      (assq type situation)))
-		     'invisible))))
+		 (memq (cdr (or (assq sym situation)
+				(assq type situation)))
+		       '(nil invisible)))))
     (setq situation (put-alist sym (if display
 				       'visible
 				     'invisible)
@@ -1884,11 +2311,6 @@ It calls function registered in variable
 
 (provide 'mime-view)
 
-(eval-when-compile
-  (setq mime-situation-examples-file nil)
-  ;; to avoid to read situation-examples-file at compile time.
-  )
-
 (mime-view-read-situation-examples-file)
 
 ;;; mime-view.el ends here
